# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    :
# @Author  :
# @Email   :
# @FileName: sql_langchain_tool.py

from typing import Type, Optional
from agentuniverse.agent.action.tool.common_tool.langchain_tool import LangChainTool
from agentuniverse.agent.action.tool.tool import ToolInput
from agentuniverse.database.sqldb_wrapper_manager import SQLDBWrapperManager
from langchain_core.tools import BaseTool, Tool as LangchainTool


class SqlLangchainTool(LangChainTool):
    db_wrapper_name: Optional[str] = ""
    clz: Type[BaseTool] = BaseTool

    def execute(self, tool_input: ToolInput):
        if self.tool is None:
            self.get_sql_database()
        return super().execute(tool_input)

    def get_sql_database(self):
        db_wrapper = SQLDBWrapperManager().get_instance_obj(self.db_wrapper_name)
        self.tool = self.clz(db=db_wrapper.sql_database)
        self.description = self.tool.description

    def as_langchain(self) -> LangchainTool:
        if self.tool is None:
            self.get_sql_database()
        return super().as_langchain()

    def get_langchain_tool(self, init_params: dict, clz: Type[BaseTool]):
        self.db_wrapper_name = init_params.get("db_wrapper")
        self.clz = clz



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/3/4 15:05
# @Author  : hiro
# @Email   : hiromesh@qq.com
# @FileName: ast_types.py

from typing import List, Any, Optional, TypedDict


class AstNodePoint(TypedDict):
    row: int
    column: int


class AstNode(TypedDict):
    type: str
    start_point: AstNodePoint
    end_point: AstNodePoint
    start_byte: int
    end_byte: int
    text: Optional[str]
    children: Optional[List['AstNode']]


class CodeBoundary(TypedDict):
    start: int
    end: int
    type: str
    name: Optional[str]
    node: Any



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/14 10:42
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: system_util.py
import ast
import inspect
import os
from pathlib import Path
import importlib
from typing import Any, Dict, List

from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum

PROJECT_ROOT_PATH = None


def get_project_root_path() -> Path:
    """Get the project root path."""
    global PROJECT_ROOT_PATH
    if PROJECT_ROOT_PATH:
        return PROJECT_ROOT_PATH
    current_work_directory = Path.cwd()
    root_path = current_work_directory.parents[1]
    PROJECT_ROOT_PATH = root_path
    return root_path


def parse_dynamic_str(param: str):
    """
    Translate a string, firstly attempting to parse it as a full path to a
    parameterless function. If it can be correctly imported, return the result
    of the function execution, otherwise return the original string.
    This is useful in scenarios where it's inconvenient to write the real
    value directly, such as with secret keys.
    """
    try:
        parts = param.rsplit('.', 1)
        if len(parts) == 2:
            module_path, func_name = parts
            module = importlib.import_module(module_path)
            func = getattr(module, func_name)
            if callable(func):
                return func()
            else:
                return param
        else:
            return param
    except (ImportError, AttributeError, Exception):
        return param


def get_module_path(yaml_path: str, root_name: str) -> str:
    """
    Convert a YAML file path to its corresponding Python module path.
    This function takes a YAML file path and a root package name, and returns the full Python module path.
    For example, if yaml_path is '/path/to/root_pkg/sub_pkg/demo_agent.yaml' and root_name is 'root_pkg',
    it will return 'root_pkg.sub_pkg.demo_agent'

    Args:
        yaml_path: The full path to the YAML file
        root_name: The name of the root package that serves as the starting point for the module path

    Returns:
        str: The full Python module path corresponding to the YAML file location

    Raises:
        FileNotFoundError: If the corresponding Python file doesn't exist
    """

    # Get the directory of the YAML file and the filename (without extension)
    dir_path = os.path.dirname(yaml_path)
    file_name = os.path.splitext(os.path.basename(yaml_path))[0]

    # Check if the corresponding Python file exists in the same directory
    py_file = os.path.join(dir_path, f"{file_name}.py")
    if not os.path.isfile(py_file):
        raise FileNotFoundError(f"Corresponding Python file not found for YAML file: {yaml_path}")

    # Get the absolute path of the directory and split it into components
    abs_dir_path = os.path.abspath(dir_path)
    path_parts = abs_dir_path.split(os.sep)

    try:
        # Find the position of the root path
        root_index = path_parts.index(root_name)
    except ValueError:
        raise FileNotFoundError(f"Corresponding Python file not found for YAML file: {yaml_path}")

    # Extract the directory parts after the root path
    relevant_dirs = path_parts[root_index + 1:]

    # Construct the module path
    module_parts = [root_name] + relevant_dirs + [file_name]
    module_path = ".".join(module_parts)

    return module_path


def process_yaml_func(func_expr: str, yaml_func_instance: Any) -> str:
    """
    Process the YAML configuration by resolving @FUNC expressions.

    Args:
        func_expr (str): The function expression to process (e.g., '@FUNC(load_api_key("qwen"))').
        yaml_func_instance (Any): The instance containing the methods to call.

    Returns:
        str: The result of the function call or the original expression if no @FUNC is found.

    Raises:
        ValueError: If @FUNC expression is provided but `yaml_func_instance` is None.
        Exception: If an error occurs while calling the method.
    """
    # Return an empty string if the function expression is None
    if not func_expr or yaml_func_instance is None:
        return func_expr

    if func_expr.startswith('@FUNC(') and yaml_func_instance is None:
        raise ValueError(f"yaml_func_extension.py is required to resolve @FUNC expression {func_expr}.")

    # Process @FUNC expressions
    if func_expr.startswith('@FUNC(') and func_expr.endswith(')'):
        # Extract the method name and arguments
        func_expr = func_expr[len('@FUNC('):-1]  # Remove '@FUNC(' and ')'
        # Parse the function call expression using AST
        try:
            parsed_expr = ast.parse(func_expr, mode='eval')
        except SyntaxError as e:
            raise ValueError(f"Invalid function expression: {func_expr}. Error: {e}")

        # Ensure the expression is a function call
        if not isinstance(parsed_expr.body, ast.Call):
            raise ValueError(f"Expected a function call, got: {func_expr}")

        # Extract function name and arguments
        func_name = parsed_expr.body.func.id
        args = [ast.literal_eval(arg) for arg in parsed_expr.body.args]
        kwargs = {kw.arg: ast.literal_eval(kw.value) for kw in parsed_expr.body.keywords}

        # Check if the method exists in the yaml function instance
        if hasattr(yaml_func_instance, func_name):
            func = getattr(yaml_func_instance, func_name)

            # Validate function signature
            sig = inspect.signature(func)
            try:
                sig.bind(*args, **kwargs)  # Ensure arguments match the function signature
            except TypeError as e:
                raise ValueError(f"Invalid arguments for method {func_name}: {e}")

            # Call the method and return the result
            try:
                res = func(*args, **kwargs)
                return res
            except Exception as e:
                raise Exception(f"Failed to execute method {func_name}: {str(e)}")
        else:
            raise AttributeError(f"Method {func_name} not found in yaml_func_extension.py")

    return func_expr


def process_dict_with_funcs(input_dict: dict, yaml_func_instance: Any) -> dict:
    """
    Process a dictionary by resolving @FUNC expressions in its values.

    Args:
        input_dict (dict): The dictionary to process.
        yaml_func_instance (Any): The instance containing the methods to call.

    Returns:
        dict: A new dictionary with resolved @FUNC expressions.

    Raises:
        ValueError: If @FUNC expression is provided but `yaml_func_instance` is None.
        Exception: If an error occurs while calling the method.
    """
    if not input_dict or yaml_func_instance is None:
        return input_dict

    processed_dict = {}
    for key, value in input_dict.items():
        if isinstance(value, str) and value.startswith('@FUNC('):
            # Process the @FUNC expression
            processed_dict[key] = process_yaml_func(value, yaml_func_instance)
        elif isinstance(value, dict):
            # Recursively process nested dictionaries
            processed_dict[key] = process_dict_with_funcs(value, yaml_func_instance)
        else:
            # Keep the value as is
            processed_dict[key] = value

    return processed_dict


def is_system_builtin(component_instance: ComponentBase) -> bool:
    """
    Determine if a component is system-built-in based on its type and configuration.

    Args:
        component_instance (ComponentBase): The component instance to check.

    Returns:
        bool: True if the component is system-built-in, False otherwise.
    """
    try:
        if component_instance is None:
            return False
        component_enum: ComponentEnum = component_instance.component_type
        if component_enum is None or not component_instance.component_config_path:
            return False

        # Define system paths for different component types
        system_paths: Dict[str, List[str]] = {
            ComponentEnum.LLM.value: ["agentuniverse", "llm", "default"],
            ComponentEnum.TOOL.value: ["agentuniverse", "agent", "action", "tool"]
        }

        # Get the system path components for this component type
        path_components = system_paths.get(component_enum.value)

        # If this component type doesn't have defined system paths, it's not a system builtin
        if not path_components:
            return False

        # Join the path components using OS-specific separator
        system_path = os.path.join(*path_components)

        # Normalize the config path to handle platform-specific separators
        normalized_config_path = os.path.normpath(component_instance.component_config_path)

        # Check if the system path is part of the normalized config path
        return system_path in normalized_config_path
    except Exception as e:
        print(f"An error occurred while checking if the component is system-built-in: {str(e)}")
        return False


def is_api_key_missing(component_instance: ComponentBase, api_key_name: str) -> bool:
    """
    Check if the given API key attribute exists and is empty.

    Args:
        component_instance (ComponentBase): The component instance to check.
        api_key_name (str): The name of the API key attribute.

    Returns:
        bool: True if the API key is missing or empty, False otherwise.
    """
    return hasattr(component_instance, api_key_name) and not getattr(component_instance, api_key_name)


def find_default_llm_config(package_list: list[str]):
    """
    Traverse the package_list to find the path containing 'agentic.llm'
    and return the absolute path of 'default_llm.toml'.

    Args:
        package_list (list[str]): The list of package paths

    Returns:
        str: The absolute path of 'default_llm.toml', or None if not found
    """
    try:
        if not package_list:
            return None

        # Iterate through each package path in the provided list
        for package_path in package_list:
            # Check if the current path contains 'agentic.llm'
            if 'agentic.llm' in package_path:
                # Try to import the module to get its actual path
                try:
                    spec = importlib.util.find_spec(package_path)
                    if spec and spec.origin:
                        # Get the directory containing the module
                        module_dir = os.path.dirname(spec.origin)
                        # Construct the full path to default_llm.toml
                        default_llm_toml_path = os.path.join(module_dir, 'default_llm.toml')

                        if os.path.exists(default_llm_toml_path):
                            return os.path.abspath(default_llm_toml_path)
                except Exception as e:
                    continue
        return None
    except Exception as e:
        print(f"An error occurred: {e}")
        return None



# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2024/12/13 11:19
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: es_conversation_memory_storage.py
import json
import datetime

import httpx
from typing import Optional, List, Any
from agentuniverse.agent.memory.conversation_memory.conversation_message import ConversationMessage
from agentuniverse.agent.memory.memory_storage.memory_storage import MemoryStorage
from agentuniverse.agent.memory.memory_storage.sql_alchemy_memory_storage import BaseMemoryConverter
from agentuniverse.agent.memory.message import Message
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger


class ElasticsearchMemoryStorage(MemoryStorage):
    """ElasticsearchMemoryStorage class that stores messages via HTTP in an Elasticsearch index.

    Attributes:
        es_url (str): The base URL of the Elasticsearch server.
        index_name (str): The name of the Elasticsearch index to store messages.
        memory_converter (BaseMemoryConverter): The memory converter to use for the memory.
    """

    es_url: Optional[str] = 'http://localhost:9200'  # The base URL of your Elasticsearch instance
    index_name: Optional[str] = 'memory'
    memory_converter: BaseMemoryConverter = None
    user: Optional[str] = None
    password: Optional[str] = None
    timeout: Optional[int] = 60
    client: Optional[httpx.Client] = None

    model_config = {
        "arbitrary_types_allowed": True,  # Allow arbitrary types
    }

    def _new_client(self):
        """Initialize the Elasticsearch HTTP client (via requests)."""
        self.client = self._client()
        self._init_es_index()

    def _initialize_by_component_configer(self,
                                          memory_storage_config: ComponentConfiger) -> 'ElasticsearchMemoryStorage':
        """Initialize the ElasticsearchMemoryStorage by the ComponentConfiger object."""
        super()._initialize_by_component_configer(memory_storage_config)
        if getattr(memory_storage_config, 'es_url', None):
            self.es_url = memory_storage_config.es_url
        if getattr(memory_storage_config, 'es_index_name', None):
            self.index_name = memory_storage_config.es_index_name
        if getattr(memory_storage_config, 'es_user', None):
            self.user = memory_storage_config.es_user
        if getattr(memory_storage_config, 'es_password', None):
            self.password = memory_storage_config.es_password
        if getattr(memory_storage_config, 'es_timeout', None):
            self.timeout = memory_storage_config.es_timeout
        if self.es_url is None:
            raise Exception('`es_url` is not set')
        # initialize the memory converter if not set
        if self.memory_converter is None:
            self.memory_converter = DefaultMemoryConverter(self.index_name)
        self._new_client()
        return self

    def _init_es_index(self):
        """Create the Elasticsearch index if it does not exist."""
        response = self.client.get(
            f'/{self.index_name}'
        )
        if response.status_code == 404:  # Index does not exist, create it
            settings = {
                "settings": {
                    "number_of_shards": 1,
                    "number_of_replicas": 0
                },
                "mappings": {
                    "properties": {
                        "session_id": {"type": "keyword"},
                        "content": {"type": "text"},
                        "trace_id": {"type": "keyword"},
                        "source": {"type": "keyword"},
                        "source_type": {"type": "keyword"},
                        "target": {"type": "keyword"},
                        "target_type": {"type": "keyword"},
                        "type": {"type": "keyword"},
                        "prefix": {"type": "text"},
                        "timestamp": {"type": "date"},
                        "params": {"type": "text"},
                        "pair_id": {"type": "keyword"},
                        "additional_args": {"type": "object"}
                    }
                }
            }
            response = self.client.put(
                f'/{self.index_name}',
                json=settings
            )
            if response.status_code != 200:
                raise Exception(f"Failed to create index: {response.text}")

    def delete(self, session_id: str = None, agent_id: str = None, trace_id: str = None, **kwargs) -> None:
        """Delete the memory from Elasticsearch.

        Args:
            session_id (str): The session id of the memory to delete.
            agent_id (str): The agent id of the memory to delete.
        """
        url = f'{self.es_url}/{self.index_name}/_delete_by_query'
        query = {
            "query": {
                "bool": {
                    "must": []
                }
            }
        }

        if session_id:
            query['query']['bool']['must'].append({"term": {"session_id": session_id}})
        if agent_id:
            query['query']['bool']['must'].append({
                "bool": {
                    "should": [
                        {"term": {"source": agent_id}},
                        {"term": {"target": agent_id}}
                    ]
                }
            })
        if trace_id:
            query['query']['bool']['must'].append({"term": {"trace_id": trace_id}})
        response = self.client.post(url, json=query)
        if response.status_code != 200:
            raise Exception(f"Failed to delete documents: {response.text}")

    def add(self, message_list: List[ConversationMessage], session_id: str = None, agent_id: str = None,
            **kwargs) -> None:
        """Add messages to the Elasticsearch index.

        Args:
            message_list (List[Message]): The list of messages to add.
            session_id (str): The session id of the memory to add.
            agent_id (str): The agent id of the memory to add.
        """
        message_list = ConversationMessage.check_and_convert_message(message_list, session_id)
        actions = []
        for message in message_list:
            action = self.memory_converter.to_es_action(message, session_id=session_id, agent_id=agent_id, **kwargs)
            actions.append(action)

        bulk_data = '\n'.join(actions) + '\n'  # Elasticsearch bulk data format requires newlines between actions
        response = self.client.post(
            f"/{self.index_name}/_bulk",
            content=bulk_data,
            headers={'Content-Type': 'application/x-ndjson'}
        )
        if response.status_code != 200:
            raise Exception(f"Failed to add documents: {response.text}")

    def get(self, session_id: str = None, agent_id: str = None, top_k=50, trace_id: str = None, **kwargs) -> List[
        ConversationMessage]:
        """Get messages from the Elasticsearch index.

        Args:
            session_id (str): The session id of the memory to get.
            agent_id (str): The agent id of the memory to get.
            top_k (int): The number of messages to get.

        Returns:
            List[Message]: The list of messages retrieved from Elasticsearch.
        """
        query = {
            "query": {
                "bool": {
                    "must": []
                }
            },
            "size": top_k,
            "sort": [
                {"timestamp": {"order": "desc"}}  # Sorting by timestamp in ascending order
            ]
        }
        if session_id:
            query['query']['bool']['must'].append({"term": {"session_id": session_id}})
        if 'type' in kwargs:
            if isinstance(kwargs['type'], list):
                memory_types = kwargs['type']
            elif isinstance(kwargs['type'], str):
                memory_types = [kwargs['type']]
            else:
                raise ValueError("type must be a list or a string")
            query['query']['bool']['must'].append({"terms": {"type": memory_types}})
        if agent_id:
            condition = {
                "bool": {
                    "must": [
                        {"match": {"target": agent_id}},
                        {"match": {"target_type": 'agent'}}
                    ]
                }
            }
            if kwargs.get('memory_types') and len(kwargs['memory_types']) > 0:
                types_condition = {
                    "bool": {
                        "must": [
                            {"match": {"source": agent_id}},
                            {"match": {"source_type": 'agent'}},
                            {"terms": {"target_type": kwargs['memory_types']}}
                        ]
                    }
                }
                query['query']['bool']['must'].append(
                    {"bool": {"should": [condition, types_condition]}})
            else:
                query['query']['bool']['must'].append(condition)

        if trace_id:
            query['query']['bool']['must'].append({"term": {"trace_id": trace_id}})
        response = self.client.post(
            f'/{self.index_name}/_search',
            json=query
        )
        if response.status_code != 200:
            raise Exception(f"Failed to retrieve documents: {response.text}")

        hits = response.json()['hits']['hits']
        messages = []
        for hit in hits:
            messages.append(self.memory_converter.from_es_hit(hit))
        messages.reverse()
        return messages

    def _client(self):
        transport = httpx.HTTPTransport(retries=3)
        if self.user and self.password:
            return httpx.Client(
                base_url=self.es_url,
                transport=transport,
                timeout=self.timeout,
                auth=(self.user, self.password),
            )
        return httpx.Client(
            base_url=self.es_url,
            transport=transport,
            timeout=self.timeout,
        )


class DefaultMemoryConverter:
    """The default memory converter for ElasticsearchMemoryStorage."""

    def __init__(self, index_name: str, **kwargs: Any):
        super().__init__(**kwargs)
        self.index_name = index_name

    def from_es_hit(self, es_hit: dict) -> Message:
        """Convert an Elasticsearch hit to a Message instance."""
        return ConversationMessage.from_dict({
            'id': es_hit['_id'],
            'conversation_id': es_hit['_source']['session_id'],
            'source': es_hit['_source']['source'],
            'source_type': es_hit['_source']['source_type'],
            'target': es_hit['_source']['target'],
            'target_type': es_hit['_source']['target_type'],
            'content': es_hit['_source']['content'],
            'metadata': {
                'prefix': es_hit['_source'].get('prefix'),
                'timestamp': datetime.datetime.fromisoformat(es_hit['_source']['timestamp']),
                'params': es_hit['_source'].get('params'),
                'pair_id': es_hit['_source'].get('pair_id'),
                'gmt_created': datetime.datetime.fromisoformat(es_hit['_source']['timestamp']).isoformat(),
            },
            'type': es_hit['_source']['type'],
            'trace_id': es_hit['_source']['trace_id'],
            'additional_args': es_hit['_source'].get('additional_args')
        })

    def to_es_action(self, message: ConversationMessage, session_id: str = None, agent_id: str = None, **kwargs) -> str:
        """Convert a message to an Elasticsearch action (index operation)."""
        document = {
            "session_id": session_id,
            "content": message.content,
            "trace_id": message.trace_id,
            "source": message.source,
            "source_type": message.source_type,
            "target": message.target,
            "target_type": message.target_type,
            "type": message.type,
            "prefix": message.metadata.get("prefix"),
            "timestamp": message.metadata.get("timestamp", datetime.datetime.now()).isoformat(),
            "params": message.metadata.get("params"),
            "pair_id": message.metadata.get("pair_id"),
            "additional_args": message.metadata.get("additional_args")

        }
        index_info = {
            "index": {"_index": self.index_name, "_id": message.id}
        }
        return f'{json.dumps(index_info)}\n{json.dumps(document)}'  # Format for Elasticsearch bulk requests



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/3/22 19:15
# @Author  : hiro
# @Email   : hiromesh@qq.com
# @FileName: write_file_tool.py

import os
import json

from agentuniverse.agent.action.tool.tool import Tool, ToolInput


class WriteFileTool(Tool):
    def execute(self, tool_input: ToolInput) -> str:
        file_path = tool_input.get_data("file_path")
        content = tool_input.get_data("content", "")
        append = tool_input.get_data("append", False)

        directory = os.path.dirname(file_path)
        if directory and not os.path.exists(directory):
            try:
                os.makedirs(directory, exist_ok=True)
            except Exception as e:
                return json.dumps({
                    "error": f"Failed to create directory: {str(e)}",
                    "file_path": file_path,
                    "status": "error"
                })

        try:
            mode = 'a' if append else 'w'

            with open(file_path, mode, encoding='utf-8') as file:
                file.write(content)
            file_size = os.path.getsize(file_path)
            return json.dumps({
                "file_path": file_path,
                "bytes_written": len(content.encode('utf-8')),
                "file_size": file_size,
                "append_mode": append,
                "status": "success"
            })

        except Exception as e:
            return json.dumps({
                "error": str(e),
                "file_path": file_path,
                "status": "error"
            })



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/3/12 23:16
# @Author  : jerry.zzw 
# @Email   : jerry.zzw@antgroup.com
# @FileName: component_configer.py
import importlib
from typing import Optional

from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.configer import Configer
from agentuniverse.base.config.custom_configer.default_llm_configer import DefaultLLMConfiger


class ComponentConfiger(object):
    """The ComponentConfiger class, which is used to load and manage the component configuration."""

    def __init__(self, configer: Optional[Configer] = None):
        """Initialize the ComponentConfiger."""
        self.__configer: Optional[Configer] = configer
        self.__metadata_type: Optional[str] = None
        self.__metadata_module: Optional[str] = None
        self.__metadata_class: Optional[str] = None
        self.__meta_class: Optional[str] = None
        self.__yaml_func_instance = None
        self.__default_llm_configer: DefaultLLMConfiger = None

    @property
    def configer(self) -> Optional[Configer]:
        """Return the Configer object."""
        return self.__configer

    @property
    def metadata_type(self) -> Optional[str]:
        """Return the type of the component."""
        return self.__metadata_type

    @property
    def metadata_module(self) -> Optional[str]:
        """Return the module of the component."""
        return self.__metadata_module

    @metadata_module.setter
    def metadata_module(self, metadata_module: str):
        self.__metadata_module = metadata_module

    @property
    def metadata_class(self) -> Optional[str]:
        """Return the class of the component."""
        return self.__metadata_class

    @metadata_class.setter
    def metadata_class(self, metadata_class: str):
        self.__metadata_class = metadata_class

    @property
    def yaml_func_instance(self):
        return self.__yaml_func_instance

    @yaml_func_instance.setter
    def yaml_func_instance(self, value):
        self.__yaml_func_instance = value

    @property
    def default_llm_configer(self) -> DefaultLLMConfiger:
        return self.__default_llm_configer

    @default_llm_configer.setter
    def default_llm_configer(self, value: DefaultLLMConfiger):
        self.__default_llm_configer = value

    @property
    def meta_class(self) -> str:
        return self.__meta_class

    def load(self) -> 'ComponentConfiger':
        """Load the configuration by the Configer object.
        Returns:
            ComponentConfiger: the ComponentConfiger object
        """
        return self.load_by_configer(self.configer)

    def load_by_configer(self, configer: Configer) -> 'ComponentConfiger':
        """Load the configuration by the Configer object.
        Args:
            configer(Configer): the Configer object
        Returns:
            ComponentConfiger: the ComponentConfiger object
        """
        self.__configer = configer

        try:
            for k, v in configer.value.items():
                self.__dict__[k] = v
            if configer.value.get('metadata'):
                self.__metadata_type = configer.value.get('metadata').get('type')
                self.__metadata_module = configer.value.get('metadata').get('module')
                self.__metadata_class = configer.value.get('metadata').get('class')
            elif configer.path and 'prompt' in configer.path:
                self.__metadata_type = ComponentEnum.PROMPT.value
            self.__meta_class = configer.value.get('meta_class')
        except Exception as e:
            raise Exception(f"Failed to parse the component configuration: {e}")

        return self

    def get_component_config_type(self) -> Optional[str]:
        """Return the type of the component.

        Returns:
            Optional[str]: the type of the component
        """
        if self.__meta_class:
            metadata_module = '.'.join(self.__meta_class.split('.')[:-1])
            metadata_class = self.__meta_class.split('.')[-1]
            module = importlib.import_module(metadata_module)
            clz = getattr(module, metadata_class)
            return clz().component_type.value
        else:
            return self.__metadata_type



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/12/9 18:01
# @Author  : fanen.lhy
# @Email   : fanen.lhy@antgroup.com
# @FileName: base_log_sink.py
from typing import Optional
from loguru import logger

from agentuniverse.base.util.logging.log_type_enum import LogTypeEnum
from agentuniverse.base.util.logging.logging_config import LoggingConfig
from agentuniverse.base.component.component_base import ComponentEnum
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.config.component_configer.component_configer import \
    ComponentConfiger



class LogSink(ComponentBase):
    """The basic class for log sink.
    """

    component_type: ComponentEnum = ComponentEnum.LOG_SINK
    name: Optional[str] = None
    description: Optional[str] = None
    level: str = "INFO"
    format: str = LoggingConfig.log_format
    sink_id: int = -1
    log_type: LogTypeEnum = LogTypeEnum.default
    enqueue: bool = True

    class Config:
        arbitrary_types_allowed = True

    def get_inheritance_depth(self):
        """
        return the depth to base Logger
        """
        return self.__class__.__mro__.index(LogSink)

    def __call__(self, message):
        self.process_record(message.record)

    def process_record(self, record):
        raise NotImplementedError("Subclasses must implement process_record.")

    def filter(self, record):
        if not record['extra'].get('log_type') == self.log_type:
            return False
        return True

    def register_sink(self):
        if self.sink_id == -1:
            self.sink_id = logger.add(
                self,
                level=self.level,
                format=self.format,
                filter=self.filter,
                enqueue=self.enqueue
            )

    def initialize_by_component_configer(self,
                                          log_sink_configer: ComponentConfiger) -> 'LogSink':
        self.name = log_sink_configer.name
        self.description = log_sink_configer.description
        if hasattr(log_sink_configer, "level"):
            self.level = log_sink_configer.level
        if hasattr(log_sink_configer, "format"):
            self.format = log_sink_configer.format
        if hasattr(log_sink_configer, "enqueue"):
            self.enqueue = log_sink_configer.enqueue

        self._initialize_by_component_configer(log_sink_configer)

        self.register_sink()
        return self



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
from typing import List, Any

from langchain_core.embeddings import Embeddings as LCEmbeddings
from pydantic import Field
from typing_extensions import Optional

from agentuniverse.agent.action.knowledge.embedding.embedding import Embedding
from agentuniverse.base.util.env_util import get_from_env
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger


# @Time : 2025/2/13 12:10
# @Author : wozhapen
# @mail : wozhapen@gmail.com
# @FileName :gemini_embedding.py

class GeminiEmbedding(Embedding):
    """Gemini Embedding class that inherits from the base Embedding class."""

    client: Any = None
    gemini_api_key: Optional[str] = Field(default_factory=lambda: get_from_env("GOOGLE_API_KEY"))

    def get_embeddings(self, texts: List[str], **kwargs) -> List[List[float]]:
        """Get embeddings for a list of texts using the Gemini API."""
        if not self.client:
            if not self.gemini_api_key:
                raise ValueError("GOOGLE_API_KEY is required but not set")
            try:
                from google import genai
                self.client = genai.Client(api_key=self.gemini_api_key)
            except ImportError as e:
                raise ImportError(
                    "genai is required. Install with: pip install google-genai"
                ) from e
            
        model_name = self.embedding_model_name or "text-embedding-004"  # default model

        try:
            response = self.client.models.embed_content(
                model=model_name,
                contents=texts,
                # gemini default 768, and only support 768
                # config=EmbedContentConfig(output_dimensionality=(self.embedding_dims or 768))
            )
            return [embedding.values for embedding in response.embeddings]
        except Exception as e:
            print(f"Error generating embedding for text: {texts}. Error: {e}")
            # Handle the error appropriately, e.g., return a zero vector or raise an exception
            raise ValueError(e)

    async def async_get_embeddings(self, texts: List[str], **kwargs) -> List[List[float]]:
        """Asynchronously get embeddings for a list of texts using the Gemini API."""
        # This implementation is synchronous because the Gemini API does not currently offer an official asynchronous client for embedding.
        # Consider implementing async batching or utilizing a ThreadPoolExecutor to wrap the synchronous call for better concurrency.
        return self.get_embeddings(texts, **kwargs)

    def as_langchain(self) -> LCEmbeddings:
        """Convert to a Langchain Embedding class."""
        #  This can be implemented to return an instance of a custom class extending from `langchain_core.embeddings.Embeddings`.
        #  However, since we are already using `google.generativeai` directly, this might be redundant.

        class GeminiLangchainEmbedding(LCEmbeddings):
            """Wrapper for Gemini Embeddings to conform to Langchain's Embeddings interface."""

            gemini_embedding: GeminiEmbedding  # Add an instance of GeminiEmbedding

            def __init__(self, gemini_embedding: GeminiEmbedding, **kwargs):
                super().__init__(**kwargs)  # Initialize the parent class
                self.gemini_embedding = gemini_embedding  # Store the GeminiEmbedding instance

            def embed_documents(self, texts: List[str]) -> List[List[float]]:
                """Embed a list of documents."""
                return self.gemini_embedding.get_embeddings(texts)

            def embed_query(self, text: str) -> List[float]:
                """Embed a single query."""
                return self.gemini_embedding.get_embeddings([text])[0]

        return GeminiLangchainEmbedding(gemini_embedding=self)  # Pass the instance of GeminiEmbedding


    def _initialize_by_component_configer(self, embedding_configer: ComponentConfiger) -> 'Embedding':
        super()._initialize_by_component_configer(embedding_configer)
        if hasattr(embedding_configer, "gemini_api_key"):
            self.gemini_api_key = embedding_configer.gemini_api_key
        
        # Initialize client if API key is available
        if self.gemini_api_key:
            try:
                from google import genai
                self.client = genai.Client(api_key=self.gemini_api_key)
            except ImportError as e:
                raise ImportError(
                    "genai is required. Install with: pip install google-genai"
                ) from e
            except Exception as e:
                raise ValueError(f"Failed to initialize Gemini client: {str(e)}")
        return self


# !/usr/bin/env python3
# -*- coding:utf-8 -*-
# @Time    : 2025/2/26 09:47
# @Author  : weizjajj 
# @Email   : weizhongjie.wzj@antgroup.com
# @FileName: openai_agent_template.py
import datetime
import json
from queue import Queue
from typing import Any, List, Dict

from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnableSerializable, RunnableConfig

from agentuniverse.agent.input_object import InputObject
from agentuniverse.agent.memory.memory import Memory
from agentuniverse.agent.memory.message import Message
from agentuniverse.agent.output_object import OutputObject
from agentuniverse.agent.plan.planner.react_planner.stream_callback import InvokeCallbackHandler, \
    OpenAIProtocolStreamOutPutCallbackHandler

from agentuniverse.agent.template.agent_template import AgentTemplate
from agentuniverse.base.context.framework_context_manager import FrameworkContextManager
from agentuniverse.base.util.prompt_util import process_llm_token
from agentuniverse.llm.llm import LLM
from agentuniverse.prompt.prompt import Prompt


class OpenAIProtocolTemplate(AgentTemplate):

    def run(self, **kwargs) -> OutputObject:
        """Agent instance running entry.

        Returns:
            OutputObject: Agent execution result
        """
        kwargs = self.parse_openai_agent_input(kwargs)
        output_object = super().run(**kwargs)
        return self.parse_openai_protocol_output(output_object)

    def parse_openai_agent_input(self, agent_input):
        for key in self.openai_protocol_input_keys():
            if key not in agent_input:
                raise ValueError(f"{key} is not in agent input")
        messages = agent_input.get('messages')
        convert_messages, image_urls = self.convert_message(messages)
        content = messages[-1].get('content')

        if isinstance(content, str):
            agent_input['input'] = content
        elif isinstance(content, list):
            for item in content:
                if item.get('type') == 'text':
                    agent_input['input'] = item.get('text')
                elif item.get('type') == 'image_url':
                    image_urls.append(item.get('image_url'))
                else:
                    raise ValueError(f"{item} is not support")
        else:
            raise ValueError(f"{content} is not support")
        if len(convert_messages) > 1:
            agent_input['chat_history'] = convert_messages[0:len(convert_messages) - 1]
        if len(image_urls) > 0:
            agent_input['image_urls'] = image_urls
        return agent_input

    def openai_protocol_input_keys(self) -> list[str]:
        return [
            'messages', 'stream'
        ]

    def convert_message(self, messages: List[Dict]):
        image_urls = []
        for message in messages:
            content = message.get('content')
            if isinstance(content, list):
                text = ""
                for item in content:
                    if item.get('type') == 'text':
                        text = item.get('text')
                    elif item.get('type') == 'image_url':
                        image_urls.append(item.get('image_url'))
                message['content'] = text
            if message.get('role') == 'user':
                message['type'] = 'human'
            elif message.get('role') == 'assistant':
                message['type'] = 'ai'
        return [Message.from_dict(message) for message in messages], image_urls


    def parse_openai_protocol_output(self, output_object: OutputObject) -> OutputObject:
        res = {
            "object": "chat.completion",
            "id": FrameworkContextManager().get_context('trace_id'),
            "created": int(datetime.datetime.now().timestamp()),
            "model": self.agent_model.info.get('name'),
            "choices": [
                {
                    "message": {
                        "role": "assistant",
                        "content": output_object.get_data('output')
                    }
                }
            ],
            "usage": {
                "prompt_tokens": 0,
                "completion_tokens": 0,
                "total_tokens": 0
            }
        }
        return OutputObject(params=res)

    def add_output_stream(self, output_stream: Queue, agent_output: str) -> None:
        if not output_stream:
            return
        output = {
            "object": "chat.completion.chunk",
            "id": FrameworkContextManager().get_context('trace_id'),
            "created": int(datetime.datetime.now().timestamp()),
            "model": self.agent_model.info.get('name'),
            "choices": [
                {
                    "delta": {
                        "role": "assistant",
                        "content": agent_output
                    },
                    "index": 0,
                }
            ]
        }
        output_stream.put(json.dumps(output))

    def customized_execute(self, input_object: InputObject, agent_input: dict, memory: Memory, llm: LLM, prompt: Prompt,
                           **kwargs) -> dict:
        self.load_memory(memory, agent_input)
        process_llm_token(llm, prompt.as_langchain(), self.agent_model.profile, agent_input)
        chain = prompt.as_langchain() | llm.as_langchain_runnable(
            self.agent_model.llm_params()) | StrOutputParser()
        res = self.invoke_chain(chain, agent_input, input_object, **kwargs)
        self.add_memory(memory, f"Human: {agent_input.get('input')}, AI: {res}", agent_input=agent_input)
        return {**agent_input, 'output': res}

    def _get_run_config(self, input_object: InputObject) -> RunnableConfig:
        config = RunnableConfig()
        callbacks = []
        output_stream = input_object.get_data('output_stream')
        callbacks.append(OpenAIProtocolStreamOutPutCallbackHandler(output_stream, agent_info=self.agent_model.info))
        callbacks.append(InvokeCallbackHandler(source=self.agent_model.info.get('name'),
                                               llm_name=self.agent_model.profile.get('llm_model').get('name')))
        config.setdefault("callbacks", callbacks)
        return config

    def invoke_chain(self, chain: RunnableSerializable[Any, str], agent_input: dict, input_object: InputObject,
                     **kwargs):
        if not self.judge_chain_stream(chain):
            res = chain.invoke(input=agent_input, config=self.get_run_config())
            return res
        result = []
        for token in chain.stream(input=agent_input, config=self.get_run_config()):
            self.add_output_stream(input_object.get_data('output_stream', None), token)
            result.append(token)
        self.add_output_stream(input_object.get_data('output_stream', None), '\n\n')
        return self.generate_result(result)



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2024/6/5 15:33
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: trace.py
import asyncio
import functools
import inspect
import time
import uuid

from functools import wraps

from agentuniverse.agent.memory.conversation_memory.conversation_memory_module import ConversationMemoryModule
from agentuniverse.base.util.monitor.monitor import Monitor
from agentuniverse.llm.llm_output import LLMOutput


def trace_llm(func):
    """Annotation: @trace_llm

    Decorator to trace the LLM invocation, add llm input and output to the monitor.
    """

    @wraps(func)
    async def wrapper_async(*args, **kwargs):
        # get llm input from arguments
        llm_input = _get_input(func, *args, **kwargs)

        source = func.__qualname__

        # check whether the tracing switch is enabled
        self = llm_input.pop('self', None)

        if self and hasattr(self, 'name'):
            name = self.name
            if name is not None:
                source = name

        if self and hasattr(self, 'tracing'):
            if self.tracing is False:
                return await func(*args, **kwargs)

        # add invocation chain to the monitor module.
        Monitor.add_invocation_chain({'source': source, 'type': 'llm'})

        start_time = time.time()
        Monitor().trace_llm_input(source=source, llm_input=llm_input)

        # invoke function
        result = await func(*args, **kwargs)
        # not streaming
        if isinstance(result, LLMOutput):
            # add llm invocation info to monitor
            Monitor().trace_llm_invocation(source=func.__qualname__, llm_input=llm_input, llm_output=result.text,
                                           cost_time=time.time() - start_time)

            # add llm token usage to monitor
            Monitor().trace_llm_token_usage(self, llm_input, result.text)
            Monitor.pop_invocation_chain()
            return result
        else:
            # streaming
            async def gen_iterator():
                llm_output = []
                async for chunk in result:
                    llm_output.append(chunk.text)
                    yield chunk
                # add llm invocation info to monitor
                output_str = "".join(llm_output)
                Monitor().trace_llm_invocation(source=func.__qualname__, llm_input=llm_input,
                                               llm_output=output_str, cost_time=time.time() - start_time)
                # add llm token usage to monitor
                Monitor().trace_llm_token_usage(self, llm_input, output_str)
                Monitor.pop_invocation_chain()

            return gen_iterator()

    @functools.wraps(func)
    def wrapper_sync(*args, **kwargs):
        # get llm input from arguments
        llm_input = _get_input(func, *args, **kwargs)

        source = func.__qualname__

        # check whether the tracing switch is enabled
        self = llm_input.pop('self', None)

        if self and hasattr(self, 'name'):
            name = self.name
            if name is not None:
                source = name

        if self and hasattr(self, 'tracing'):
            if self.tracing is False:
                return func(*args, **kwargs)

        # add invocation chain to the monitor module.
        Monitor.add_invocation_chain({'source': source, 'type': 'llm'})

        start_time = time.time()
        Monitor().trace_llm_input(source=source, llm_input=llm_input)

        # invoke function
        result = func(*args, **kwargs)
        # not streaming
        if isinstance(result, LLMOutput):
            # add llm invocation info to monitor
            Monitor().trace_llm_invocation(source=source, llm_input=llm_input, llm_output=result.text,
                                           cost_time=time.time() - start_time)

            # add llm token usage to monitor
            Monitor().trace_llm_token_usage(self, llm_input, result.text)
            Monitor.pop_invocation_chain()
            return result
        else:
            # streaming
            def gen_iterator():
                llm_output = []
                for chunk in result:
                    llm_output.append(chunk.text)
                    yield chunk
                # add llm invocation info to monitor
                output_str = "".join(llm_output)
                Monitor().trace_llm_invocation(source=func.__qualname__, llm_input=llm_input,
                                               llm_output=output_str, cost_time=time.time() - start_time)

                # add llm token usage to monitor
                Monitor().trace_llm_token_usage(self, llm_input, output_str)
                Monitor.pop_invocation_chain()

            return gen_iterator()

    if asyncio.iscoroutinefunction(func):
        # async function
        return wrapper_async
    else:
        # sync function
        return wrapper_sync


def get_caller_info(instance: object = None):
    source_list = Monitor.get_invocation_chain()
    if len(source_list) > 0:
        return {
            'source': source_list[-1].get('source'),
            'type': source_list[-1].get('type')
        }
    else:
        return {
            'source': '',
            'type': 'user'
        }


def trace_agent(func):
    """Annotation: @trace_agent

    Decorator to trace the agent invocation, add agent input and output to the monitor.
    """

    @functools.wraps(func)
    async def wrapper_async(*args, **kwargs):
        # get agent input from arguments
        agent_input = _get_input(func, *args, **kwargs)
        # check whether the tracing switch is enabled
        source = func.__qualname__
        self = agent_input.pop('self', None)

        tracing = None
        if isinstance(self, object):
            agent_model = getattr(self, 'agent_model', None)
            if isinstance(agent_model, object):
                info = getattr(agent_model, 'info', None)
                profile = getattr(agent_model, 'profile', None)
                if isinstance(info, dict):
                    source = info.get('name', None)
                if isinstance(profile, dict):
                    tracing = profile.get('tracing', None)
        start_info = get_caller_info()
        pair_id = f"agent_{uuid.uuid4().hex}"
        kwargs['memory_source_info'] = start_info
        ConversationMemoryModule().add_agent_input_info(start_info, self, agent_input, pair_id)
        if tracing is False:
            result = await func(*args, **kwargs)
            ConversationMemoryModule().add_agent_result_info(self, result, start_info, pair_id)
            return result

        # add invocation chain to the monitor module.
        Monitor.init_invocation_chain()
        Monitor.add_invocation_chain({'source': source, 'type': 'agent'})

        start_time = time.time()
        Monitor().trace_agent_input(source=source, agent_input=agent_input)

        # invoke function
        result = await func(*args, **kwargs)
        # add agent invocation info to monitor
        Monitor().trace_agent_invocation(source=source, agent_input=agent_input, agent_output=result,
                                         cost_time=time.time() - start_time)
        ConversationMemoryModule().add_agent_result_info(self, result, start_info, pair_id)
        Monitor.pop_invocation_chain()
        return result

    @functools.wraps(func)
    def wrapper_sync(*args, **kwargs):
        # get agent input from arguments
        agent_input = _get_input(func, *args, **kwargs)
        # check whether the tracing switch is enabled
        source = func.__qualname__
        self = agent_input.pop('self', None)

        tracing = None
        if isinstance(self, object):
            agent_model = getattr(self, 'agent_model', None)
            if isinstance(agent_model, object):
                info = getattr(agent_model, 'info', None)
                profile = getattr(agent_model, 'profile', None)
                if isinstance(info, dict):
                    source = info.get('name', None)
                if isinstance(profile, dict):
                    tracing = profile.get('tracing', None)
        pair_id = f"agent_{uuid.uuid4().hex}"
        start_info = get_caller_info()
        kwargs['memory_source_info'] = start_info
        ConversationMemoryModule().add_agent_input_info(start_info, self, agent_input, pair_id)
        if tracing is False:
            result = func(*args, **kwargs)
            ConversationMemoryModule().add_agent_result_info(self, result, start_info, pair_id)
            return result

        # add invocation chain to the monitor module.
        Monitor.init_invocation_chain()
        Monitor.add_invocation_chain({'source': source, 'type': 'agent'})

        start_time = time.time()
        Monitor().trace_agent_input(source=source, agent_input=agent_input)

        # invoke function
        result = func(*args, **kwargs)
        # add agent invocation info to monitor
        Monitor().trace_agent_invocation(source=source, agent_input=agent_input, agent_output=result,
                                         cost_time=time.time() - start_time)
        ConversationMemoryModule().add_agent_result_info(self, result, start_info, pair_id)
        Monitor.pop_invocation_chain()
        return result

    if asyncio.iscoroutinefunction(func):
        # async function
        return wrapper_async
    else:
        # sync function
        return wrapper_sync


def trace_tool(func):
    """Annotation: @trace_tool

    A decorator to trace tool invocations, supporting both synchronous and asynchronous functions.
    It monitors tool execution, tracks timing, and maintains an invocation chain.
    """

    def process_tool(source, tool_input, start_info, pair_id):
        """Process common tool logic

        Args:
            source: The source/name of the tool
            tool_input: Input parameters for the tool
            start_info: Information about where the tool was called from
            pair_id: The ID for this tool invocation

        Returns:
            tuple: (self tool instance, updated source name)
        """
        self = tool_input.pop('self', None)
        if isinstance(self, object):
            name = getattr(self, 'name', None)
            if name is not None:
                source = name
        ConversationMemoryModule().add_tool_input_info(start_info, source, tool_input, pair_id)
        return self, source

    def handle_tool_result(start_info, source, result, pair_id):
        """Handle the tool execution result

        Args:
            start_info: Information about where the tool was called from
            source: The source/name of the tool
            result: The execution result
            pair_id: The ID for this tool invocation

        Returns:
            The execution result
        """
        ConversationMemoryModule().add_tool_output_info(start_info, source, params=result, pair_id=pair_id)
        return result

    def trace_tool_execution(source, tool_input, result, start_time):
        """Trace the tool execution process

         Args:
             source: The source/name of the tool
             tool_input: Input parameters for the tool
             result: The execution result
             start_time: Timestamp when the tool execution started
         """
        # add invocation chain to the monitor module.
        Monitor.add_invocation_chain({'source': source, 'type': 'tool'})
        Monitor().trace_tool_invocation(
            source=source,
            tool_input=tool_input,
            tool_output=result,
            cost_time=time.time() - start_time
        )
        Monitor.pop_invocation_chain()

    @functools.wraps(func)
    async def wrapper_async(*args, **kwargs):
        # Extract tool input from arguments
        tool_input = _get_input(func, *args, **kwargs)
        start_time = time.time()
        source = func.__qualname__
        start_info = get_caller_info()
        pair_id = f"tool_{uuid.uuid4().hex}"

        self, source = process_tool(source, tool_input, start_info, pair_id)

        # Handle case when tracing is disabled
        if self and hasattr(self, 'tracing') and self.tracing is False:
            result = await func(*args, **kwargs)
            return handle_tool_result(start_info, source, result, pair_id)

        # Initialize and execute with full tracing
        Monitor.init_invocation_chain()
        Monitor().trace_tool_input(source, tool_input)
        result = await func(*args, **kwargs)
        trace_tool_execution(source, tool_input, result, start_time)
        return handle_tool_result(start_info, source, result, pair_id)

    @functools.wraps(func)
    def wrapper_sync(*args, **kwargs):
        # Extract tool input from arguments
        tool_input = _get_input(func, *args, **kwargs)
        start_time = time.time()
        source = func.__qualname__
        start_info = get_caller_info()
        pair_id = f"tool_{uuid.uuid4().hex}"

        self, source = process_tool(source, tool_input, start_info, pair_id)

        # Handle case when tracing is disabled
        if self and hasattr(self, 'tracing') and self.tracing is False:
            result = func(*args, **kwargs)
            return handle_tool_result(start_info, source, result, pair_id)

        # Initialize and execute with full tracing
        Monitor.init_invocation_chain()
        Monitor().trace_tool_input(source, tool_input)
        result = func(*args, **kwargs)
        trace_tool_execution(source, tool_input, result, start_time)
        return handle_tool_result(start_info, source, result, pair_id)

    return wrapper_async if asyncio.iscoroutinefunction(func) else wrapper_sync


def trace_knowledge(func):
    """Annotation: @trace_knowledge

    Decorator to trace the knowledge invocation.
    """

    def process_knowledge(source, knowledge_input, start_info, pair_id):
        """Process common knowledge logic

        Args:
            source: The source/name of the knowledge
            knowledge_input: Input parameters for the knowledge
            start_info: Information about where the knowledge was called from
            pair_id: The ID for this knowledge invocation

        Returns:
            tuple: (self knowledge instance, updated source name)
        """
        self = knowledge_input.pop('self', None)
        if isinstance(self, object):
            name = getattr(self, 'name', None)
            if name is not None:
                source = name
        ConversationMemoryModule().add_knowledge_input_info(start_info, source, knowledge_input, pair_id)
        return self, source

    def handle_knowledge_result(start_info, source, result, pair_id):
        """Handle the knowledge execution result

        Args:
            start_info: Information about where the knowledge was called from
            source: The source/name of the knowledge
            result: The execution result
            pair_id: The ID for this knowledge invocation

        Returns:
            The execution result
        """
        ConversationMemoryModule().add_knowledge_output_info(start_info, source, params=result, pair_id=pair_id)
        return result

    def trace_knowledge_execution(source):
        """Trace the knowledge execution process

        Args:
            source: The source/name of the knowledge
        """
        Monitor.init_invocation_chain()
        Monitor.add_invocation_chain({'source': source, 'type': 'knowledge'})

    @functools.wraps(func)
    async def wrapper_async(*args, **kwargs):
        # Get knowledge input from arguments
        knowledge_input = _get_input(func, *args, **kwargs)
        source = func.__qualname__
        start_info = get_caller_info()
        pair_id = f"knowledge_{uuid.uuid4().hex}"

        self, source = process_knowledge(source, knowledge_input, start_info, pair_id)

        # Handle case when tracing is disabled
        if self and hasattr(self, 'tracing') and self.tracing is False:
            result = await func(*args, **kwargs)
            return handle_knowledge_result(start_info, source, result, pair_id)

        # Initialize and execute with full tracing
        trace_knowledge_execution(source)
        result = await func(*args, **kwargs)
        Monitor.pop_invocation_chain()
        return handle_knowledge_result(start_info, source, result, pair_id)

    @functools.wraps(func)
    def wrapper_sync(*args, **kwargs):
        # Get knowledge input from arguments
        knowledge_input = _get_input(func, *args, **kwargs)
        source = func.__qualname__
        start_info = get_caller_info()
        pair_id = f"knowledge_{uuid.uuid4().hex}"

        self, source = process_knowledge(source, knowledge_input, start_info, pair_id)

        # Handle case when tracing is disabled
        if self and hasattr(self, 'tracing') and self.tracing is False:
            result = func(*args, **kwargs)
            return handle_knowledge_result(start_info, source, result, pair_id)

        # Initialize and execute with full tracing
        trace_knowledge_execution(source)
        result = func(*args, **kwargs)
        Monitor.pop_invocation_chain()
        return handle_knowledge_result(start_info, source, result, pair_id)

    return wrapper_async if asyncio.iscoroutinefunction(func) else wrapper_sync


def _get_input(func, *args, **kwargs) -> dict:
    """Get the agent input from arguments."""
    sig = inspect.signature(func)
    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()
    return {k: v for k, v in bound_args.arguments.items()}



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/3/16 15:28
# @Author  : wangchongshi
# @Email   : wangchongshi.wcs@antgroup.com
# @FileName: llm_channel.py
from typing import Optional, Any, Union, Iterator, AsyncIterator

import httpx
import openai
import tiktoken
from openai import OpenAI, AsyncOpenAI
from langchain_core.language_models import BaseLanguageModel

from agentuniverse.agent.memory.message import Message
from agentuniverse.base.annotation.trace import trace_llm
from agentuniverse.base.component.component_base import ComponentBase
from agentuniverse.base.component.component_enum import ComponentEnum
from agentuniverse.base.config.application_configer.application_config_manager import ApplicationConfigManager
from agentuniverse.base.config.component_configer.component_configer import ComponentConfiger
from agentuniverse.llm.llm_channel.langchain_instance.default_channel_langchain_instance import \
    DefaultChannelLangchainInstance
from agentuniverse.llm.llm_output import LLMOutput


class LLMChannel(ComponentBase):
    channel_name: Optional[str] = None
    channel_api_key: Optional[str] = None
    channel_api_base: Optional[str] = None
    channel_organization: Optional[str] = None
    channel_proxy: Optional[str] = None
    channel_model_name: Optional[str] = None
    channel_ext_info: Optional[dict] = None

    model_support_stream: Optional[bool] = None
    model_support_max_context_length: Optional[int] = None
    model_support_max_tokens: Optional[int] = None
    model_is_openai_protocol_compatible: Optional[bool] = True

    _channel_model_config: Optional[dict] = None
    client: Any = None
    async_client: Any = None
    component_type: ComponentEnum = ComponentEnum.LLM_CHANNEL

    def _initialize_by_component_configer(self, component_configer: ComponentConfiger) -> 'LLMChannel':

        super()._initialize_by_component_configer(component_configer)
        if hasattr(component_configer, "channel_name"):
            self.channel_name = component_configer.channel_name
        if hasattr(component_configer, "channel_api_key"):
            self.channel_api_key = component_configer.channel_api_key
        if hasattr(component_configer, "channel_api_base"):
            self.channel_api_base = component_configer.channel_api_base
        if hasattr(component_configer, "channel_organization"):
            self.channel_organization = component_configer.channel_organization
        if hasattr(component_configer, "channel_proxy"):
            self.channel_proxy = component_configer.channel_proxy
        if hasattr(component_configer, "channel_model_name"):
            self.channel_model_name = component_configer.channel_model_name
        if hasattr(component_configer, "channel_ext_info"):
            self.channel_ext_info = component_configer.channel_ext_info
        if hasattr(component_configer, "model_support_stream"):
            self.model_support_stream = component_configer.model_support_stream
        if hasattr(component_configer, "model_support_max_context_length"):
            self.model_support_max_context_length = component_configer.model_support_max_context_length
        if hasattr(component_configer, "model_support_max_tokens"):
            self.model_support_max_tokens = component_configer.model_support_max_tokens
        if hasattr(component_configer, "model_is_openai_protocol_compatible"):
            self.model_is_openai_protocol_compatible = component_configer.model_is_openai_protocol_compatible
        return self

    def create_copy(self):
        return self

    @property
    def channel_model_config(self):
        return self._channel_model_config

    @channel_model_config.setter
    def channel_model_config(self, config: dict):
        self._channel_model_config = config
        if config:
            for key, value in config.items():
                if not isinstance(key, str):
                    continue
                if key == 'streaming':
                    if self.model_support_stream is False:
                        value = False
                if key == 'max_tokens':
                    if self.model_support_max_tokens:
                        value = min(self.model_support_max_tokens, value) if value else self.model_support_max_tokens
                if key == 'max_context_length':
                    if self.model_support_max_context_length:
                        value = min(self.model_support_max_context_length,
                                    value) if value else self.model_support_max_context_length
                if not self.__dict__.get(key):
                    self.__dict__[key] = value

    @trace_llm
    def call(self, *args: Any, **kwargs: Any):
        """Run the LLM."""
        return self._call(*args, **kwargs)

    @trace_llm
    async def acall(self, *args: Any, **kwargs: Any):
        """Asynchronously run the LLM."""
        return self._acall(*args, **kwargs)

    def _call(self, messages: list, **kwargs: Any) -> Union[LLMOutput, Iterator[LLMOutput]]:
        streaming = kwargs.pop("streaming") if "streaming" in kwargs else self.channel_model_config.get('streaming')
        if 'stream' in kwargs:
            streaming = kwargs.pop('stream')
        if self.model_support_stream is False and streaming is True:
            streaming = False

        support_max_tokens = self.model_support_max_tokens
        max_tokens = kwargs.pop('max_tokens', None) or self.channel_model_config.get('max_tokens', None) or support_max_tokens
        if support_max_tokens:
            max_tokens = min(support_max_tokens, max_tokens)

        self.client = self._new_client()
        chat_completion = self.client.chat.completions.create(
            messages=messages,
            model=kwargs.pop('model', self.channel_model_name),
            temperature=kwargs.pop('temperature', self.channel_model_config.get('temperature')),
            stream=kwargs.pop('stream', streaming),
            max_tokens=max_tokens,
            **kwargs,
        )
        if not streaming:
            text = chat_completion.choices[0].message.content
            return LLMOutput(text=text, raw=chat_completion.model_dump(),
                             message=Message(content=chat_completion.choices[0].message.content,
                                             type=chat_completion.choices[0].message.role))
        return self.generate_stream_result(chat_completion)

    async def _acall(self, messages: list, **kwargs: Any) -> Union[LLMOutput, AsyncIterator[LLMOutput]]:
        streaming = kwargs.pop("streaming") if "streaming" in kwargs else self.channel_model_config.get('streaming')
        if 'stream' in kwargs:
            streaming = kwargs.pop('stream')
        if self.model_support_stream is False and streaming is True:
            streaming = False

        support_max_tokens = self.model_support_max_tokens
        max_tokens = kwargs.pop('max_tokens', None) or self.channel_model_config.get('max_tokens', None) or support_max_tokens
        if support_max_tokens:
            max_tokens = min(support_max_tokens, max_tokens)

        self.async_client = self._new_async_client()
        chat_completion = await self.async_client.chat.completions.create(
            messages=messages,
            model=kwargs.pop('model', self.channel_model_name),
            temperature=kwargs.pop('temperature', self.channel_model_config.get('temperature')),
            stream=kwargs.pop('stream', streaming),
            max_tokens=max_tokens,
            **kwargs,
        )
        if not streaming:
            text = chat_completion.choices[0].message.content
            return LLMOutput(text=text, raw=chat_completion.model_dump(),
                             message=Message(content=chat_completion.choices[0].message.content,
                                             type=chat_completion.choices[0].message.role))
        return self.agenerate_stream_result(chat_completion)

    def as_langchain(self) -> BaseLanguageModel:
        """Convert to the langchain llm class."""
        return DefaultChannelLangchainInstance(self)

    def get_num_tokens(self, text: str) -> int:
        """Get the number of tokens present in the text.

        Useful for checking if an input will fit in an openai model's context window.

        Args:
            text: The string input to tokenize.

        Returns:
            The integer number of tokens in the text.
        """

        encoding = tiktoken.get_encoding("cl100k_base")
        return len(encoding.encode(text))

    def max_context_length(self) -> int:
        return self.channel_model_config.get('max_context_length')

    def _new_client(self):
        """Initialize the openai client."""
        if self.client is not None:
            return self.client
        return OpenAI(
            api_key=self.channel_api_key,
            organization=self.channel_organization,
            base_url=self.channel_api_base,
            timeout=self.channel_model_config.get('request_timeout'),
            max_retries=self.channel_model_config.get('max_retries'),
            http_client=httpx.Client(proxy=self.channel_proxy) if self.channel_proxy else None,
            **(self.channel_model_config.get('client_args') or {}),
        )

    def _new_async_client(self):
        """Initialize the openai async client."""
        if self.async_client is not None:
            return self.async_client
        return AsyncOpenAI(
            api_key=self.channel_api_key,
            organization=self.channel_organization,
            base_url=self.channel_api_base,
            timeout=self.channel_model_config.get('request_timeout'),
            max_retries=self.channel_model_config.get('max_retries'),
            http_client=httpx.AsyncClient(proxy=self.channel_proxy) if self.channel_proxy else None,
            **(self.channel_model_config.get('client_args') or {}),
        )

    def generate_stream_result(self, stream: openai.Stream):
        """Generate the result of the stream."""
        for chunk in stream:
            llm_output = self.parse_result(chunk)
            if llm_output:
                yield llm_output

    async def agenerate_stream_result(self, stream: AsyncIterator) -> AsyncIterator[LLMOutput]:
        """Generate the result of the stream."""
        async for chunk in stream:
            llm_output = self.parse_result(chunk)
            if llm_output:
                yield llm_output

    @staticmethod
    def parse_result(chunk):
        """Generate the result of the stream."""
        chat_completion = chunk
        if not isinstance(chunk, dict):
            chunk = chunk.dict()
        if len(chunk["choices"]) == 0:
            return
        choice = chunk["choices"][0]
        message = choice.get("delta")
        text = message.get("content")
        role = message.get("role")
        if text is None:
            text = ""
        return LLMOutput(text=text, raw=chat_completion.model_dump(), message=Message(content=text, type=role))

    def get_instance_code(self) -> str:
        """Return the full name of the component."""
        appname = ApplicationConfigManager().app_configer.base_info_appname
        return f'{appname}.{self.component_type.value.lower()}.{self.channel_name}'



# !/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    :
# @Author  :
# @Email   :
# @FileName: search_api_tool.py

from typing import Optional

from agentuniverse.agent.action.tool.tool import Tool, ToolInput
from agentuniverse.base.config.component_configer.configers.tool_configer import ToolConfiger
from agentuniverse.base.util.env_util import get_from_env
from langchain_community.utilities import SearchApiAPIWrapper
from pydantic import Field


class SearchAPITool(Tool):
    """
    The demo search tool.

    Implement the execute method of demo google search tool, using the `SearchApiAPIWrapper` to implement a simple search.

    Note:
        You need to sign up for a free account at https://www.searchapi.io/ and get the SEARCHAPI_API_KEY api key (100 free queries).

    Args:
        search_api_key: Optional[str] = Field(default_factory=lambda: get_from_env("SEARCHAPI_API_KEY")),
        engine: str = "google" engine type you want to use
        search_params: dict = {} engine search parameters
        search_type: str = "common" result type you want to get ,common string or json
    """

    search_api_key: Optional[str] = Field(default_factory=lambda: get_from_env("SEARCHAPI_API_KEY"))
    engine: str = "google"
    search_params: dict = {}
    search_api_wrapper: Optional[SearchApiAPIWrapper] = None
    search_type: str = "common"

    def _load_api_wapper(self):
        if not self.search_api_key:
            raise ValueError("Please set the SEARCHAPI_API_KEY environment variable.")
        if not self.search_api_wrapper:
            self.search_api_wrapper = SearchApiAPIWrapper(searchapi_api_key=self.search_api_key, engine=self.engine)
        return self.search_api_wrapper

    def execute(self, tool_input: ToolInput):
        self._load_api_wapper()
        search_params = {}
        for k, v in self.search_params.items():
            if k in tool_input.to_dict():
                search_params[k] = tool_input.get_data(k)
                continue
            search_params[k] = v
        input = tool_input.get_data("input")
        if self.search_type == "json":
            return self.search_api_wrapper.results(query=input, **search_params)
        return self.search_api_wrapper.run(query=input, **search_params)

    def initialize_by_component_configer(self, component_configer: ToolConfiger) -> 'Tool':
        """Initialize the tool by the component configer."""
        super().initialize_by_component_configer(component_configer)
        self.engine = component_configer.configer.value.get('engine', 'google')
        self.search_params = component_configer.configer.value.get('search_params', {})
        self.search_type = component_configer.configer.value.get('search_type', 'common')
        return self
